/*
 *  Copyright (C) 2014  Thadeu Lima de Souza Cascardo <cascardo@cascardo.eti.br>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

static int dump(int fd, size_t len, int pgsz, char *fname)
{
	int out;
	char *page;
	int i, npages;
	int r;
	page = malloc(pgsz);
	if (!page)
		return -ENOMEM;
	out = open(fname, O_WRONLY | O_CREAT | O_EXCL, 0644);
	if (out < 0)
		return -errno;
	npages = (len + pgsz - 1) / pgsz;
	for (i = 1; i <= npages; i++) {
		if (i == npages) {
			len = len & (pgsz - 1);
			if (len > 0) {
				r = read(fd, page, len);
				if (r != len)
					goto out;
				r = write(out, page, len);
				if (r != len)
					goto out;
			}
		} else {
			r = read(fd, page, pgsz);
			if (r != pgsz)
				goto out;
			r = write(out, page, pgsz);
			if (r != pgsz)
				goto out;
		}
	}
	free(page);
	close(out);
	return 0;
out:
	free(page);
	close(out);
	if (r < 0)
		return -errno;
	return -EINVAL;
}

int main(int argc, char **argv)
{
	int fd;
	uint32_t kernel, initrd, second, pgsz;
	char buffer[4096];
	char *p;
	int r;
	off_t offset;
	if (argc < 2)
		return 1;
	fd = open(argv[1], O_RDONLY);
	if (fd < 0) {
		return 2;
	}
	r = read(fd, buffer, sizeof(buffer));
	if (r < 32)
		return 3;
	p = buffer + 8;
	if (strncmp(buffer, "ANDROID!", 8)) {
		return 4;
	}
	kernel = *(uint32_t *) p;
	p += 8;
	initrd = *(uint32_t *) p;
	p += 8;
	second = *(uint32_t *) p;
	p += 12;
	pgsz = *(uint32_t *) p;
	offset = pgsz;
	if (kernel > 0) {
		lseek(fd, offset, SEEK_SET);
		dump(fd, kernel, pgsz, "kernel");
	}
	offset += ((kernel + pgsz - 1) / pgsz) * pgsz;
	if (initrd > 0) {
		lseek(fd, offset, SEEK_SET);
		dump(fd, initrd, pgsz, "initrd");
	}
	offset += ((initrd + pgsz - 1) / pgsz) * pgsz;
	if (second > 0) {
		lseek(fd, offset, SEEK_SET);
		dump(fd, second, pgsz, "second");
	}
	close(fd);
	return 0;
}
